import math

def TeX_code(value):
    '''
    Return TeX code for float value as power of 10

    Input:
    value : numerical value to express as power of 10, e.g. 1e-3

    Returns:
        value to express as power of 10, e.g. 1e-3 as $10^{-3}$

    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com
    
    Created on Mar 8, 2023
    '''
    
    value = abs(value)
    if value > 1.0:
        power = round(math.log10(value))
        return '$10^{' + str(power) + '}$'
    elif value < 1.0:
        power = round(math.log10(value))
        return '$10^{' + str(power) + '}$'
    else:
        return '$10^{0}$'



'''
Gets list of Comma, Separated, Value (CSV) files in a directory

Input:

directory: directory path containing desired list of files

Output:

csv_files: a list of CSV files in the the directory

Created on Jun 28, 2023

Author: Kevin Wu
        Parsons Corporation
        www.parsons.com
'''
import os

def get_files_in_directory(directory):
    if not os.path.exists(directory):
        raise ValueError("Directory does not exist: " + directory)
    csv_files = []
    for file in os.listdir(directory):
        extension = os.path.splitext(file)[1]
        extension = extension.lower()
        if extension == '.csv':
            csv_files.append(file)
    return csv_files


'''
This function compares the csv files in a pair of directories. It first gets the model directory and 
checks to see if the files in the model directory exist in the reference directory. It the checks if 
any column statistic has a fail and if it does it will mark all_success as 'Fail'

Input:
mod_dir: the model directory the user want to test
ref_dir: reference directory the user want to compare the model directory to

Output:
file_dict: a dictionary with the key being the file name and the value being the statistic
all_success: the value that checks to see whether the test has passed or failed
Created on Jun 28, 2023

Author: Kevin Wu
        Parsons Corporation
        www.parsons.com
'''

import os
from opendss_scivis.get_file_directory import get_files_in_directory
from opendss_scivis.colname_stats import colname_stats

def file_stats(mod_dir, ref_dir):
    mod_file=get_files_in_directory(mod_dir)
    all_success = 'Pass' # Assume test for all files passes by default
    file_dict = {}
    for file_name in mod_file:
        ref_file_path = os.path.join(ref_dir, file_name)
        if not os.path.isfile(ref_file_path):
            raise Exception(f"File '{file_name}' does not exist in the reference directory.")
        mod_file_path = os.path.join(mod_dir, file_name)
        col_dict=colname_stats(mod_file_path,ref_file_path)
        file_dict[file_name]=col_dict
        if col_dict['success'] != 'Pass':
            all_success='Fail'
    return file_dict,all_success


    


def is_list_of_strings(lst):
    '''
    Check if list contains only strings.

    Input:
    lst : list that may be all strings

    Returns:
      true if all strings
      false if not all strings

    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com
    
    Created on Apr 22, 2022
    '''
    
    if lst and isinstance(lst, list):
        return all(isinstance(elem, str) for elem in lst)
    else:
        return False



from matplotlib.ticker import ScalarFormatter
import matplotlib.pyplot as plt
from matplotlib import rcParams
import matplotlib

def plot_xy_axes(axes,option,subplot_axis):
    '''
    Plot axes for X-Y plot.
    
    Plots the x & y axes for a X-Y plot using the information provided in 
    the AXES dictionary returned by the get_time_series_axes function.
    
    INPUTS:
    axes   : dictionary containing axes information for X-Y plot
    option  : dictionary containing option values. (Refer to 
              get_time_series_plot_options function for more information.)
    option['xlabel'] : label to use for x-axis
    option['ylabel'] : label to use for y-axis
    subplot_axis : axis handle for subplot
    
    OUTPUTS:
    None
    
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com

    Created on May 1, 2022
    '''
    
    # ax = plt.gca()
    index_x = option['index_x']
    index_y = option['index_y']
    nlabel = len(option['axeslabel'])

    # Set new ticks
    plt.xticks(axes['xtick'])
    plt.yticks(axes['ytick'])

    # Set axes limits
    axislim = [axes['xtick'][0], axes['xtick'][-1], axes['ytick'][0], axes['ytick'][-1]]
    plt.axis(axislim)
    
    plottype = option['plottype']
    if plottype == 'linear':

        # Check whether to scale an axis to scientific notation
        if option['axismax'][index_y-1] > 100.0:
            # Set formatter of the major ticker
            ax = plt.gca()
            yScalarFormatter = ScalarFormatterClass(useMathText=True)
            yScalarFormatter.set_powerlimits((0,0))
            ax.yaxis.set_major_formatter(yScalarFormatter)
        
    # Label x-axis
    fontSize = matplotlib.rcParams.get('font.size')
    xpos = axes['xtick'][-1] + 2*axes['xtick'][-1]/30
    ypos = axes['xtick'][-1]/30
    if nlabel > 0:
        if index_x >=0 and index_x < nlabel:
            xlabelh = subplot_axis.set_xlabel(option['axeslabel'][index_x], fontsize = fontSize, \
                                 horizontalalignment = 'center')
        else:
            raise ValueError("option['axeslabel'][" + str(index_x) + "] does not exist")

    # Label y-axis
    xpos = 0
    ypos = axes['ytick'][-1] + 2*axes['ytick'][-1]/30
    if nlabel > 0:
        if index_x >=0 and index_x < nlabel:
            ylabelh = subplot_axis.set_ylabel(option['axeslabel'][index_y], fontsize = fontSize, \
                                 horizontalalignment = 'center')
        else:
            raise ValueError("option['axeslabel'][" + str(index_x) + "] does not exist")

    subplot_axis.tick_params(axis='both', direction='in', which='both') # have ticks inside plot
    
    # Set axes line width
    lineWidth = rcParams.get('lines.linewidth')
    subplot_axis.spines['left'].set_linewidth(lineWidth)
    subplot_axis.spines['right'].set_linewidth(lineWidth)
    subplot_axis.spines['top'].set_linewidth(lineWidth)
    subplot_axis.spines['bottom'].set_linewidth(lineWidth)

class ScalarFormatterClass(ScalarFormatter):
   def _set_format(self):
       # Specify only 1 significant digit after decimal place
       self.format = "%1.1f"



import os
import pandas as pd

def read_element_data(name):
    '''
    Read data of an element in power grid that contains a time series of monitored power 
    system quantities from a Comma Separated Value (CSV) file output by OpenDSS.

    Input:
    name : name of CSV file with or without suffix, e.g. SimpleDemo.csv or SimpleDemo. 

    Output:
    objectData : data contained in CSV file as a two-dimensional data structure with labeled axes
    
    The data format of objectData is a DataFrame as provided by the pandas.read_csv function. An 
    example of the format for the contents of an OpenDSS CVS file is 
    
          hour   t(sec)   Frequency  ...        PShaft  dSpeed (Deg/sec)  dTheta (Deg)
0        0    0.001     60.0000  ...  2.000000e+09         -0.000093 -9.000000e-10
1        0    0.002     60.0000  ...  2.000000e+09         -0.000093 -2.430000e-09
2        0    0.003     60.0000  ...  2.000000e+09         -0.000093 -4.040000e-09
3        0    0.004     60.0000  ...  2.000000e+09         -0.000093 -5.660000e-09
4        0    0.005     60.0000  ...  2.000000e+09         -0.000093 -7.280000e-09
...    ...      ...         ...  ...           ...               ...           ...
4979     0    4.980     60.4106  ...  2.000000e+09       -575.187000  2.579670e+00
4980     0    4.981     60.4090  ...  2.000000e+09       -576.661000  2.569620e+00
4981     0    4.982     60.4074  ...  2.000000e+09       -578.108000  2.559550e+00
4982     0    4.983     60.4058  ...  2.000000e+09       -579.527000  2.549440e+00
4983     0    4.984     60.4041  ...  2.000000e+09       -580.918000  2.539320e+00

[4984 rows x 8 columns]

    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com
    
    Created on Apr 22, 2022
    '''

    # Check if CSV file suffix
    file_name, file_extension = os.path.splitext(name)
    if file_extension == "":
        filename = name + '.csv'
    elif file_extension.lower()=='.csv':
        filename = name
    else:
        raise Exception("Invalid file type: " + name)
    
    # Check if file exists and is readable
    if not os.path.isfile(filename):
        raise Exception("File does not exist: " + filename)
    elif not os.access(filename, os.R_OK):
        raise Exception("File is not readable: " + filename)
        
    # Load object from CSV file
    try:
        # Default of comma delimiter
        objectData = pd.read_csv(filename)
    except:
        # Second case of tab delimiter
        objectData = pd.read_csv(filename, delimiter='\t')
    
    return objectData



'''
This script was created to facilitate identification of undesired side effects
of updates related to the generation of target and Taylor diagrams using the
existing examples.

It supports the following argument as an option.

-clean_files : Delete new files if this flag is present

It can be invoked from a command line as:

$ python test_plots.py -clean_files

The script supports the argument "-h" for help (call "$ python test_plots.py -h").

The python scripts executed are those specified by SCRIPTS_TESTED below that
reside in the "Examples" folder. These scripts are executed without displaying
the plots but write new example graphics files. The new files are compared
pixel-wise with the previously existing example graphic files and a percentage
change is displayed for each example.

Percentage change: 100% * sum of pixel color differences / mean colors intensity

Some image files are generated with different width/height depending on the system.
When the size of the current image and the size of the new image are different,
both are resized to have the same size and then compared. This is reported in the
script output if it occurs.

Changes less than 2% may be considered negligible (GOOD). Between 2% and 10% should be
double checked (CHECK). More than 10% is a probable case of a difference arsing due to a
bug (BAD). The test_plot script reports these outcomes according to the percentage.
An example of the script output is shown below.

$ python3 test_plots.py
Executing 9 scripts in sequence:
 01:   0.86% distance between 'target1_example.png' and 'target1.png', GOOD.
 02:   0.15% distance between 'target2_example.png' and 'target2.png', GOOD.
 03:   0.15% distance between 'target3_example.png' and 'target3.png', GOOD.
 04:   0.00% distance between 'target4_example.png' and 'target4.png', GOOD.
 05:   0.00% distance between 'target5_example.png' and 'target5.png', GOOD.
 06:   0.11% distance between 'target6_example.png' and 'target6.png', GOOD.
 07:   0.00% distance between 'target7_example.png' and 'target7.png', GOOD.
 08:   2.64% distance between 'target8_example.png' and 'target8.png', CHECK (size adjusted).
 09:   2.48% distance between 'target9_example.png' and 'target9.png', CHECK (size adjusted).
New files were kept after comparison.


The list of scripts to be processed must be assigned to SCRIPTS_TESTED as a list of
tuples, and can make use of regular expressions (regex) for filename pattern
matching that is supported by the glob module. For example to process all
target and taylor scripts:

SCRIPTS_TESTED = ("target*[0-9].py", "taylor*[0-9].py")

To process only the target or taylor scripts:

SCRIPTS_TESTED = ("target*[0-9].py", ) # target diagrams only, must be a tuple

SCRIPTS_TESTED = ("taylor*[0-9].py", ) # Taylor diagrams only, must be a tuple

The list of files are processed in sorted alphabetical order.

Author: Andre D. L. Zanchetta
        ("adlzanchetta" in multiple social media)

Created on Aug 28, 2022

@author: adlzanchetta@gmail.com
'''

from PIL import Image
import numpy as np
import subprocess
import argparse
import glob
import os

import pandas  # this script does not use pandas, but import it to ensure the
               #    functionality of the others

# ## CONSTANTS ################################################################## #

IMAGES_FORMAT = ".png"
EXAMPLES_FOLDER_NAME = 'Examples'
SCRIPTS_TESTED = ("example*[0-9].py", )
PYTHON_COMMAND = "python <SCRIPT> -noshow"
DEBUG_FILE_NAME = "<BASE>%s" % IMAGES_FORMAT
EXAMP_FILE_NAME = "<BASE>_ref%s" % IMAGES_FORMAT

# ## DEFS ####################################################################### #

def change_cwd() -> None:
    """
    Move the current working directory (cwd) to the Examples folder
    """

    test_folder_path = os.path.dirname(os.path.realpath(__file__))
    base_folder_path = os.path.split(test_folder_path)[0]
    examples_folder_path = os.path.join(base_folder_path, EXAMPLES_FOLDER_NAME)
    os.chdir(examples_folder_path)

    return None


def get_scripts_to_run() -> tuple:
    """
    List scripts that produce plots
    :return: List of script files paths that match all Regexs in SCRIPTS_TESTED
    """

    ret_list = []
    for tested_glob in SCRIPTS_TESTED:
        ret_list += glob.glob(tested_glob)
        del tested_glob
    return tuple(sorted(ret_list))


def get_ndarrays(pillow_img) -> np.ndarray:
    """
    Gets the ndarray of a pillow image prone to differences in pillow version
    """

    try:
        # for Python 3.6.5, pillow 8.4.0
        return pillow_img.__array__()
    except AttributeError as e:
        # for Python 3.10.4, pillow 9.2.0
        return np.array(pillow_img)


def get_comparable_pixels_rgba(file_path_1: str, file_path_2: str) -> tuple:
    """
    Read image files and get their data in the same shape (width x height)
    :return: (nd array with RGBA of arg 1, nd array with RGBA of arg2, bool if
                images were resized)
    """

    # open image files and get 2D dimensions
    img_prev, img_curr = Image.open(file_path_1), Image.open(file_path_2)
    pixels_prev, pixels_curr = get_ndarrays(img_prev), get_ndarrays(img_curr)
    shape_prev, shape_curr = pixels_prev.shape, pixels_curr.shape

    # files have equal size then nothing needs to be done
    if sum([1 if p != c else 0 for p, c in zip(shape_prev, shape_curr)]) == 0:
        img_prev.close(), img_curr.close()
        return pixels_prev, pixels_curr, False

    # get the sizes of the smallest
    min_x = min(shape_prev[0], shape_curr[0])
    min_y = min(shape_prev[1], shape_curr[1])
    
    # reshape both images
    min_xy = (min_y, min_x)
    img_prev, img_curr = img_prev.resize(min_xy), img_curr.resize(min_xy)

    # crop both images
    # box_xy = (0, 0, min_y, min_x)
    # img_prev, img_curr = img_prev.crop(box_xy), img_curr.crop(box_xy)

    # get and close
    pixels_prev, pixels_curr = get_ndarrays(img_prev), get_ndarrays(img_curr)
    img_prev.close(), img_curr.close()

    return pixels_prev, pixels_curr, True


def compare_rasters_pixelwise(prev: str, curr: str) -> tuple:
    """
    Calculates the distance between rasters in their color space (RGB, RGBA, BW...)
    :param prev: Previous version image file path
    :param curr: Current version image file path
    :return: Percentual color distance between images, boolean flag if images were
                resized if comparison was possible, raise errors otherwise
    """

    try:
        pixels_prev, pixels_curr, resized = get_comparable_pixels_rgba(prev, curr)

        # get images 'intensities' (color magnitude)
        pixels_maginitude_1, pixels_maginitude_2 = np.sum(pixels_prev), np.sum(pixels_curr)
        pixels_maginitude_mean = np.mean([pixels_maginitude_1, pixels_maginitude_2])
        del pixels_maginitude_1, pixels_maginitude_2

        # get pixels percentual distance
        pixels_dist = np.sum(pixels_prev - pixels_curr)
        pixels_dist_pct = 100 * pixels_dist / pixels_maginitude_mean
        del pixels_dist, pixels_maginitude_mean
        
        return pixels_dist_pct, resized
        
    except FileNotFoundError as e:
        print("  ", e)
        raise e


def evaluate_output(all_script_file_names: tuple, clean_files: bool) -> None:
    """
    Runs sequence of examples, comparing outputs and printing findings in to STDOUT
    :param all_script_file_names: Sequence of scripts to be executed.
    :param clean_files: If true, remove new files after comparison. Keep them otherwise.
    :return: None
    """

    print("Executing %d scripts in sequence:" % len(all_script_file_names))

    for script_count, script_file_name in enumerate(all_script_file_names):

        # get new and old filenames
        output_file_name = script_file_name[0:-3]
        output_file_name_new = DEBUG_FILE_NAME.replace("<BASE>", output_file_name)
        output_file_name_old = EXAMP_FILE_NAME.replace("<BASE>", output_file_name)
        del output_file_name    

        # define command to execute
        py_command = PYTHON_COMMAND.replace("<SCRIPT>", script_file_name)

        if os.path.exists(output_file_name_old):
            
            # run the command and calculate files distance
            subprocess.run(py_command, shell=True, stdout=subprocess.DEVNULL)
            rasters_distance_pct, resized = compare_rasters_pixelwise(
                output_file_name_old, output_file_name_new)
            
            # communicate result
            if rasters_distance_pct < 2:
                # Changes lower than 2% may be considered negligible
                status = 'GOOD'
            elif rasters_distance_pct < 10:
                # Changes between 2% and 10% should be checked
                status = 'CHECK'
            else:
                # More than 10% is a probable case of bug created
                status = 'BAD'
            print(" %02d: %6.02f%% distance between '%s' and '%s', %s%s." %
                (script_count+1, rasters_distance_pct, output_file_name_old,
                 output_file_name_new, status, " (size adjusted)" if resized else ""))

            del rasters_distance_pct, resized

        else:
            # if old image file does not exist, skip
            print(" %02d. Skipping '%s': base figure not found." % 
                (script_count+1, script_file_name))
        
        # delete new file if needed
        if clean_files and os.path.exists(output_file_name_new):
            os.remove(output_file_name_new)

        del output_file_name_new, output_file_name_old
        del script_count, script_file_name

    return None


# ## MAIN ####################################################################### #

if __name__ == '__main__':

    # Defines the output file name or path 
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-clean_files', dest='clean_files', action='store_true',
                            help="Delete new files if this flag is present.")
    args = arg_parser.parse_args()
    del arg_parser

    # Move to the Examples folder
    change_cwd()

    # list scripts to execute
    test_script_file_names = get_scripts_to_run()

    # execute each one
    evaluate_output(test_script_file_names, args.clean_files)

    # wrapping up message
    new_files_were = "deleted" if args.clean_files else "kept"
    print("New files were %s after comparison." % new_files_were)


from opendss_scivis import get_time_series_axes
from opendss_scivis import get_time_series_plot_options
from opendss_scivis import plot_xy
from opendss_scivis import plot_xy_axes

import math
import matplotlib.pyplot as plt

def time_series_plot(*args, **kwargs):
    '''
    Plot a time series of one or more quantities calculated in OpenDSS.
    
    time_series_plot(data,keyword=value)
    
    The first arguments must be the input as described below followed by
    keywords in the format OPTION = value. An example call to the function 
    would be:
    
    ToDo: update as package evolves
    time_series_plot(variables,data,markerdisplayed='marker')
    
    INPUTS:
    data : dictionary returned from get_time_series function

    OUTPUTS:
    None.
    
    LIST OF OPTIONS:
    For an exhaustive list of options to customize your plot, call the 
    function without arguments at a Python command line:
    % python
    >>> import opendss_scivis as osv
    >>> osv.time_series_plot()

    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com
    
    Created on Apr 22, 2022
    '''
    global number_cols, number_rows, subplot_axis

    nargin = len(args)
    data = _get_time_series_plot_arguments(*args,**kwargs)
    if nargin == 0: return
    
    # Get options
    # option = get_time_series_plot_options(data,**kwargs)
    option = get_time_series_plot_options(**kwargs)
    #  Get time axis values for plot
    key_list = list(data)
    key = key_list[0]
    time = data[key].get('values')
    
    # Get number of time series to plot
    max_number_plots = len(data.keys()) - 1
    nlabel = len(option['axeslabel'])
    if nlabel == 0:
        units = data[key].get('units')
        if len(units) > 0:
            axis_label = key + ' (' + units + ')'
        else:
            axis_label = key
        option['axeslabel'].append(axis_label)
    
    # Do maximum of 6 subplots per page
    for iplot in range(0, max_number_plots, 6):
        page = int(iplot/6)
        number_plots = min(max_number_plots - 6*page,6)
        option['nplots'] = number_plots
        option['index_x'] = 0

        # Set matrix of subplots for figure
        fig = _set_subplots(option)
     
        # Produce subplots
        for x in range(1,number_plots+1):
            # Get quantity
            key = key_list[iplot+x]
            quantity = data[key].get('values')
    
            option['index_y'] = iplot+x
            if nlabel == 0: 
                units = data[key].get('units')
                if len(units) > 0:
                    axis_label = key + ' (' + units + ')'
                else:
                    axis_label = key
                    
                option['axeslabel'].append(axis_label)
            
            # Get axes values for plot
            axes = get_time_series_axes(time,quantity,option)
    
            # Produce x-y plot
            plot_xy(time,quantity,option,subplot_axis[x-1])
    
            # Modify axes for this plot
            if option['overlay'] == 'off':
            
                # Plot axes
                plot_xy_axes(axes,option,subplot_axis[x-1])
            else:
                markerLabel = option['markerlabel']
                subplot_axis[x-1].legend(markerLabel)
        
        # Hide last subplot if odd number of subplots in 2 columns
        number_rows = math.ceil(number_plots / number_cols)
        # Updated code to handle the box in the last plot
        if number_cols == 2:
            if number_plots % 2 == 1:
                subplot_axis[-1].axis('off')
                if number_rows > 1:
                    subplot_index = (number_rows - 1) * number_cols
                    subplot_axis[subplot_index].spines['top'].set_visible(True)
                    subplot_axis[subplot_index].spines['right'].set_visible(True)
                    subplot_axis[subplot_index].spines['bottom'].set_visible(True)
                    subplot_axis[subplot_index].spines['left'].set_visible(True)
       
def _get_time_series_plot_arguments(*args,**kwargs):
    '''
    Get arguments for time_series_plot function.
    
    Retrieves the arguments supplied to the time_series_plot function as
    arguments and displays the optional arguments if none are supplied.
    Otherwise, tests the first argument is a non-empty dictionary and returns it.
        
    INPUTS:
    args : variable-length input argument list
    
    OUTPUTS:
    time_series_data: time series of quantities in a dictionary
    '''
    
    time_series_data = {}
    nargin = len(args)
    if nargin == 0:
        # Display options list
        _display_time_series_plot_options()
        return time_series_data
    elif nargin != 1:
        raise ValueError('Must supply 1 argument')

    time_series_data = args[0]

    # Test the data type of first argument
    if not isinstance(time_series_data, dict):
        raise ValueError('First argument is not a dictionary')
    elif not time_series_data:
        raise ValueError('First argument is an empty dictionary')

    return time_series_data

def _display_time_series_plot_options():
    '''
    Displays available options for TAYLOR_DIAGRAM function.
    '''
    
    _disp('General options:')
    _dispopt("'colormap'","'on'/ 'off' (default): "  + 
        "Switch to map color shading of markers to colormap ('on')\n\t\t"  +
        "or min to max range of RMSDz values ('off').")
    _dispopt("'overlay'","'on' / 'off' (default): " +
        'Switch to overlay current statistics on Taylor diagram. ' +
        '\n\t\tOnly markers will be displayed.')
    _disp('')
    
    _disp('Marker options:')
    _dispopt("'MarkerDisplayed'",
        "'line' : use a continuous line (default)\n\t\t"  + 
        "'marker' : disp0lay markers at provide points")
    _disp("OPTIONS when 'MarkerDisplayed' == 'line'")
    _dispopt("'lineSpec'","Line specification (default solid black, 'k-')")
    _dispopt("'lineWidth'","Line width specification (default default rcParams " + 
             "'lines.linewidth'")

    _disp("OPTIONS when 'MarkerDisplayed' == 'marker'")
    _dispopt("'alpha'","Blending of symbol face color (0.0 transparent through 1.0 opaque)" +
             "\n\t\t" + "(Default: 1.0)")
    _dispopt("'markerColor'",'Single color to use for all markers'  +
        ' (Default: red)')
    _dispopt("'markerLabel'",'Labels for markers. (Default: None)')
    _dispopt("'markerLabelColor'",'Marker label color (Default: black)')
    _dispopt("'markerLegend'","'on' / 'off' (default): "  +
        'Use legend for markers')
    _dispopt("'markers'",'Dictionary providing individual control of the marker label, ' + 
             'label color, \n\t\t' + 
             'symbol, size, face color, and edge color. (Default: none)')
    _dispopt("'markerSize'",'Marker size (Default: 10)')
    _dispopt("'markerSymbol'","Marker symbol (Default: '.')")
    _disp('')
    
    _disp('Axes options:')
    _dispopt("'axislim'",'Axes limits [xmin, xmax, ymin, ymax] (Default: data values)')
    _dispopt("'ticks'",'Define tick positions (default is that used by axis function)')
    _dispopt("'xtickLabelPos'",'Position of the tick labels along the x-axis ' + 
             '(empty by default)')
    _dispopt("'ytickLabelPos'",'Position of the tick labels along the y-axis ' + 
             '(empty by default)')
    _disp('')
    
    _disp('Plot options:')
    _dispopt("'plotType'",
        "'linear': Standard linear plot\n\t\t"  + 
        "'loglog': Log scaling on both the x and y axis\n\t\t"  + 
        "'semilogx': Log scaling on the x axis\n\t\t"  + 
        "'semilogy' : Log scaling on the y axis")

def _disp(text):
    print(text)

def _dispopt(optname,optval):
    '''
    Displays option name and values

    This is a support function for the display_time_series_plot_options function.
    It displays the option name OPTNAME on a line by itself followed by its 
    value OPTVAL on the following line.
    '''

    _disp('\t%s' % optname)
    _disp('\t\t%s' % optval)

def _set_subplots(option):
    global number_cols, number_rows, subplot_axis

    number_plots = option['nplots']
    number_cols = option['layout'][1]

    # Constrain number of columns to within number of plots
    if number_cols > number_plots: number_cols = number_plots

    if option['overlay'] == 'off':
        number_rows = math.ceil(number_plots / number_cols)
        
        if number_plots < 4:
            fig, ax = plt.subplots(number_rows, number_cols)
        elif number_plots < 7:
            fig, ax = plt.subplots(number_rows, number_cols, figsize=(8.5, 11))
        else:
            raise ValueError('Not programmed to handle more than 6 subplots.')

        if number_plots == 1:
            subplot_axis = []
            subplot_axis.append(ax)
        else:
            subplot_axis = ax.flatten()

        return fig



import pandas as pd
from opendss_scivis.is_list_of_strings import is_list_of_strings

def get_time_series(variables, CSVdata, **kwargs):
    '''
    Extract time series data from a data structure obtained from a Comma Separated Value (CSV) 
    file output by OpenDSS.

    Input:
    variables : a string array of time series variables to be extracted as specified in the 
                column headers of the CSV file.
                If an empty list, then all time series variables are extracted
    data      : time series in a two-dimensional data structure with labeled axes
    *kwargs : variable-length keyword argument list. ??The keywords by 
              definition are dictionaries with keys that must correspond to 
              one choices given in OUTPUTS below.
    
    The format of data is a DataFrame as provided by the pandas.read_csv function. 

    Output:
    timeSeriesData : extracted time series in a dictionary with each variable in its own separate
                dictionary

    Each time series dictionary is of the form:

    CSV_variable = {"values": timeSeries, "units": units}
    
    where
    
    values : is a list containing the date time series
    units  : is the units for the variable
    
    For example, 
        {'values': [0.001, 0.002, 0.003, 0.004, 0.005, 0.006, ... ], 'units': 'sec'}
    
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com
    
    Created on Apr 25, 2022
    '''

    # Set default parameters for all options
    option = {}
    option['subsample'] = 0

    # Check data types of input variables
    
    # Test the data types of first two arguments
    if not isinstance(variables, list):
        raise ValueError('First argument is not a list')
    elif not isinstance(CSVdata, pd.DataFrame):
        raise ValueError('Second argument is not a Pandas DataFrame')
    
    # Check variables is list of strings
    if variables and not is_list_of_strings(variables):
        raise ValueError('First argument is not a string list')
    
    # Check for valid keys and values in dictionary
    for optname, optvalue in kwargs.items():
        optname = optname.lower()
        if not optname in option:
            raise ValueError('Unrecognized option: ' + optname)
        else:
            # Replace option value with that from arguments
            option[optname] = optvalue

    # Find out contents of data
    labels = []
    for col in CSVdata.columns:
        labels.append(col.strip())
    
    timeSeriesData = {}
    if variables:
        # Extract requested data and return in a dictionary
        for name in variables:
            if name in labels:
                index = labels.index(name)
                timeSeries = CSVdata.iloc[:,index].tolist()
                if option['subsample'] > 0:
                    step = option['subsample']
                    timeSeries = timeSeries[0:-1:step]
                
                # Extract units if present and name dictionary to just variable
                start = name.find('(')
                end = name.find(')')
                if start != -1 and end != -1:
                    CSV_variable = name[0:start].strip()
                    units = name[start+1:end]
                else:
                    CSV_variable = name.strip()
                    units = ""
                
                # Create dictionary with value and units
                timeSeriesData[CSV_variable] = {"values": timeSeries, "units": units}
            else:
                raise ValueError('Variable ' + name + ' not in data.')
    else:
        # Extract all data and return in a dictionary
        index = 0
        while index < len(labels):
            name = labels[index]
            if name.find('Unnamed') != -1:
                # skip any unnamed columns
                break
            if name == 'hour':
                # Add hour and t(sec) to just t(sec)
                hour = CSVdata.iloc[:,index].tolist()
                index += 1
                time = CSVdata.iloc[:,index].tolist()
                timeSeries = []
                for x in range(0,len(hour)):
                    timeSeries.append(3600*hour[x] + time[x])
            else:
                timeSeries = CSVdata.iloc[:,index].tolist()
            
            # Extract units if present and name dictionary to just variable
            start = name.find('(')
            end = name.find(')')
            if start != -1 and end != -1:
                CSV_variable = name[0:start].strip()
                units = name[start+1:end]
            else:
                CSV_variable = name.strip()
                units = ""
            
            # Create dictionary with value and units
            timeSeriesData[CSV_variable] = {"values": timeSeries, "units": units}
            index += 1

    return timeSeriesData



import numpy as np
from opendss_scivis.read_element_data import read_element_data
from skill_metrics import bias
from skill_metrics import bias_percent
from skill_metrics import centered_rms_dev
from skill_metrics import rmsd

'''
This file allows users to compare files to see if there is any difference between them. 

It first sets a tolerance, for limiting the amount of difference, then counts to see how 
many differences are in the file. The function also searches/creates a folder for the 
comparison file. Then it takes in the Comma Separated Value (CSV) files the user wants 
to compare and checks to see if the files are the same size. Finally we compare
the CSV files to see if the two files have any significant differences based on the 
tolerance set.

Input:
file_old: name of the reference for the CSV file is trying to compare
file_new: name of the CSV file user wants to compare

output:
resolve: a dictionary containing ... (need to provide details once function finished)

Author: Kevin Wu
        Parsons Corporation
        www.parsons.com

Created on Jun 7, 2023

@author: kevin.wu@xatorcorp.com
'''
def colname_stats(file_ref, file_mod):
    '''
    Calculate statistical metrics of the differences between Comma Separated Value (CSV) files
    for each column in the file.
    
    The function first checks to see if the dimensions of the two CSV files are the same. 
    It then loops through each column and checks if the columns in the two files are the same. 
    If it is not the same, then it will calculate several statistical metrics and return them in 
    a dictionary which will be the value of the outer dictionary with the key of the outer 
    dictionary the column name. 
    
    Example output: {col1:{'bias':bias,'crmsd':crmsd,...}, ...
                     col2:{'bias':bias,'crmsd':crmsd,...},...}
    
    Input:
    file_ref: the reference CSV file
    file_mod: the model CSV file that is used to compare with the reference
    
    Output: 
    col_dict: a dictionary that contains dictionaries of statistical metrics
    '''
    
    # Get data from CSV files
    csv1 = read_element_data(file_ref)
    csv2 = read_element_data(file_mod)
    # if csv1.shape == csv2.shape:
    #     print("The files have the same size")
    # else:
    #     raise Exception("The file sizes do not match")

    '''
    Create a dictionary where the key is the column name and its values is another
    dictionary containing the statistical metrics calculated, e.g., 
    metrics = {"bias": bias, "rmse": rmse, ...}. 

    This provides the flexibility to add other metrics returned by the function in future
    without breaking earlier written code. 
    '''
    
    file_success = 'Pass' # Assume test for all column passes by default
    
    col_dict={}
    for col in csv1.columns:
        if col == 'Unnamed: 8':
            break
        metrics={}
        if not csv1[col].equals(csv2[col]):
            value_ref=np.array(csv1[col].tolist())
            value_model=np.array(csv2[col].tolist())
            metrics['bias']= bias(value_model,value_ref)
            metrics['crmsd']=centered_rms_dev(value_model,value_ref)
            metrics['rmsd']=rmsd(value_model,value_ref)
    
            sdevm = np.std(value_model)
            sdevr = np.std(value_ref)
            metrics['sdev'] = [sdevr, sdevm]
            
            if sdevm == 0 or sdevr == 0:
                ccoef = -1
            else:
                ccoef=np.corrcoef(value_model,value_ref)
                metrics['ccoef'] = ccoef[0]

            Bp = bias_percent(value_model,value_ref)
            metrics['bias_percent'] = Bp
            
            metrics['success'] = pass_fail(metrics)
            if metrics['success']=='Fail':
                file_success='Fail'
        else:
            metrics['success'] = 'Pass'

        col_dict[col] = metrics
    
    col_dict['success'] = file_success

    return col_dict

'''
ToDo: We'll need to investigate using other statistical metrics for the pass/fail. 
The t-test is based on the assumption one is comparing two statistical samples for
a significant difference. For our current application, we are doing a test to see
if a software application exactly reproduces a previous result, i.e. a verification 
test. We expect the element by element difference to be zero or relatively small and
wish to know whenever such is the case. 
'''
def pass_fail(metrics,tolerance=0.1):
    '''
    This function checks the quotient of the bias and the reference average and compares it 
    with a tolerance which determines if there is significant difference.
    
    Input:
    m:     model data
    r:     reference data
    tolerance: the value that determines whether to pass or fail
    
    Output:
    Pass: if percentage is less than tolerance
    Fail: if percentage is greater than tolerance
    '''
    #ToDo: check for presence of key 'bias_percent'
    if metrics['bias_percent'] < tolerance:
        return 'Pass'
    else:
        return 'Fail'



import numpy as np
import pandas as pd
import os
import sys
sys.path.append('C:\Users\FF16GK3\OpenDSS_SciVis')
import scipy.stats as stats
from skill_metrics import centered_rms_dev
from OpenDSS_SciVis.opendss_scivis.read_element_data import read_element_data


'''
This file allows users to compare files to see if there is any difference between them. 

It first sets a tolerance, for limiting the amount of difference, then counts to see how 
many differences are in the file. The function also searches/creates a folder for the 
comparison file. Then it takes in the Comma Separated Value (CSV) files the user wants 
to compare and checks to see if the files are the same size. Finally we compare
the CSV files to see if the two files have any significant differences based on the 
tolerance set.

Input:
file_old: name of the reference for the CSV file is trying to compare
file_new: name of the CSV file user wants to compare

output:
resolve: a dictionary containing ... (need to provide details once function finished)

Author: Kevin Wu
        Parsons Corporation
        www.parsons.com

Created on Jun 7, 2023

@author: kevin.wu@xatorcorp.com
'''
def colname_stat_dict(file_ref, file_mod,stat_list=[]):
    '''
    This is used to create a dictionary that contains another dictionary. 
    
    It first checks if stat_list has any value. Then it checks to see if the dimensions of the two csv files
    are the same. It loops through each column and checks if each column are the same. If it is not the same
    it will solve all the statistical metrics and put it into a dictionary which will be the value of the 
    outer dictionary and the key of the outer dictionary is the column name
    
    Example output: {col1:{'bias':bias,'crmsd':crmsd,...},col2:{{'bias':bias,'crmsd':crmsd,...},...}
    
    Input:
    file_ref: the reference csv file
    file_mod: the model csv file that is used to compare with the reference
    stat_list: a list for the statistical metrics that need to be solved
    
    Output: 
    col_dict: a dictionary that contains dictionaries of statistical metrics
    '''
    if len(stat_list)==0:
        stat_list=['bias_percent','bias','crmsd','rmsd','ccoef','sdev']
    
    csv1 = read_element_data(file_ref)
    csv2 = read_element_data(file_mod)
    if csv1.shape == csv2.shape:
        print("The files have the same size")
    else:
        raise Exception("The files size do not match")

    #ToDo: Have this function return a dictionary where the key is the 
    # column name, e.g. "P1" in the case of "SimpleDemo_Mon_g1_1.csv". 
    # different_columns['P1'] = 
    # Each element associated with the key will be another dictionary
    # containing the statistical metrics calculated, e.g., 
    # metrics = {"bias": bias, "rmse": rmse}
    # different_columns['P1'] = metrics
    #
    # We will essentially be returning a dictionary of dictionaries. This provides
    # the flexibility to add other metrics returned by the function in future
    # without breaking earlier written code. 
    #
    # Also put this function into a separate file as part of the opendss_scivis module. 
    
    col_dict={}
    for col in csv1.columns:
        if not csv1[col].equals(csv2[col]):
            stat_dict={}
            value_ref=np.array(csv1[col].tolist())
            value_model=np.array(csv2[col].tolist())
            for stat in stat_list:
                if stat == 'bias':
                    stat_dict['bias']= np.mean(value_model) - np.mean(value_ref)
                elif stat == 'crmsd':
                    #converts list to array because the function requires ndarray
                    stat_dict['crmsd']=centered_rms_dev(value_model,value_ref)
                elif stat == 'rmsd':
                    stat_dict['rmsd']=np.sqrt(np.sum(np.square(np.subtract(value_model,value_ref)))/float(value_model.size))
                elif stat == 'ccoef':
                    ccoef=np.corrcoef(value_model,value_ref)
                    stat_dict['ccoef'] = ccoef[0]
                elif stat == 'sdev':
                    sdevm = np.std(value_model)
                    sdevr = np.std(value_ref)
                    stat_dict['sdev'] = [sdevr, sdevm]
                elif stat == 'bias_percent':
                    #(model-reference)/reference
                    stat_dict['bias_percent']=pass_fail(value_ref, value_model)
            col_dict[col]=stat_dict
    print(col_dict)
    return col_dict

#ToDo: For now, leave this as a helper function for colname_difference. 
def calc_stat_by_col(file_ref, file_mod):
    '''
    This function takes the list from the colname_difference function and 
    converts the values in each column of the list. Using the list to find the average 
    and bias of each column.
    
    Input:
    file_old: the reference csv file
    file_new: the model csv file that is used to compare with the reference
    
    Output:
    stats: a dictionary containing statistics of the differences
    stats['bias'] = bias
    ...
    '''
    column_list = colname_stat_dict(file_ref, file_mod)
    csv1 = read_element_data(file_ref)
    csv2 = read_element_data(file_mod)
    stats={}
    for name in column_list:
        value1=csv1[name].tolist()
        value2=csv2[name].tolist()

        #ToDo: use statistics functions from the Skill Metrics package where 
        # possible, e.g. for the bias. This will ensure consistency with any 
        # target and Taylor diagrams we create with the same data. 
        avg1=np.mean(value1)
        avg2=np.mean(value2)
        bias= avg1-avg2
        stats[name]={"bias": bias}
    print(stats)
    return stats

'''
ToDo: We'll need to investigate using other statistical metrics for the pass/fail. 
The t-test is based on the assumption one is comparing two statistical samples for
a significant difference. For our current application, we are doing a test to see
if a software application exactly reproduces a previous result, i.e. a verification 
test. We expect the element by element difference to be zero or relatively small and
wish to know whenever such is the case. 
'''
def pass_fail(m,r,tolerance=0.01):
    '''
    This function checks the quotient of the bias and the reference average and compares it 
    with a tolerance which determines if there is significant difference.
    
    Input:
    m:     model data
    r:     reference data
    tolerance: the value that determines whether to pass or fail
    
    Output:
    Pass: if percentage is less than tolerance
    Fail: if percentage is greater than tolerance
    '''
    model=np.mean(m)
    ref=np.mean(r)
    bias_percent=abs((model-ref)/ref)
    if bias_percent<tolerance:
        return 'Pass'
    else:
        return 'Fail'

#ToDo: This needs to be one of the quantities returned in the metrics dictionary 
# of as discussed above in my feedback on the colname_difference function.
# We may also want to give some thought on renaming the functions but this should
# be relatively easy to do. (I implemented this in the colname_stat_dict function)
def csv_file_col_variation(file_ref, file_mod,tolerance=0.01):
    '''
    This function takes the list from the colname_difference function and 
    converts the values in each column of the list. Using the list to find the 
    average and bias of each column.
    
    Input:
    file_old: the reference csv file
    file_new: the model csv file that is used to compare with the reference
    
    Output:
    bias_dict: a dictionary that hold all the bias value from the columns 
    '''
    column_list = colname_stat_dict(file_ref, file_mod)
    csv1 = read_element_data(file_ref)
    csv2 = read_element_data(file_mod)
    test_result={}
    for name in column_list:
        value1=csv1[name].tolist()
        value2=csv2[name].tolist()
        pass_fail_result=pass_fail(value2, value1)
        test_result[name]=pass_fail_result
    print(test_result)
    return test_result


__version__='1.1.0'

from .add_line_legend import add_line_legend
from .add_marker_legend import add_marker_legend
from .colname_stats import colname_stats
from .compare_directory import file_stats
from .get_file_directory import get_files_in_directory
from .get_time_series import get_time_series
from .get_time_series_axes import get_time_series_axes
from .get_time_series_plot_options import get_time_series_plot_options
from .is_list_of_strings import is_list_of_strings
from .plot_xy import plot_xy
from .plot_xy_axes import plot_xy_axes
from .plot_xy_line import plot_xy_line
from .plot_xy_markers import plot_xy_markers
from .read_element_data import read_element_data
from .TeX_code import TeX_code
from .time_series_plot import time_series_plot



import matplotlib.pyplot as plt
import matplotlib.colors as clr
import matplotlib
import warnings
from opendss_scivis import add_marker_legend

def plot_xy_markers(X,Y,option):
    '''
    Plots a one-dimensional variable in an x-y plot using markers.
    
    Plots a function Y(X) such as frequency with time.
    
    INPUTS:
    x       : independent coordinates
    y       : function values at x-coordinates
    option  : dictionary containing option values. (Refer to 
              get_time_series_plot_options function for more information.)
    option['axismax'] : maximum for the X & Y values. Used to limit maximum 
                        range to display markers
    option['axismin'] : minimum for the X & Y values. Used to limit minimum 
                        range to display markers
    option['index_x'] : index for x values in option
    option['index_y'] : index for y values in option

    OUTPUTS:
    None
    
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com

    Created on May 1, 2022
    '''
    
    index_x = option['index_x']
    index_y = option['index_y']

    # Set face color transparency
    alpha = option['alpha']
    
    # Set font and marker size
    fontSize = matplotlib.rcParams.get('font.size') - 2
    markerSize = option['markersize']
    
    if option['markerlegend'] == 'on':
        # Check that marker labels have been provided
        if option['markerlabel'] == '':
            raise ValueError('No marker labels provided.')

        # Plot markers of different color and shapes with labels 
        # displayed in a legend
        
        # Define markers
        kind = ['+','o','x','s','d','^','v','p','h','*']
        colorm = ['b','r','g','c','m','y','k']
        if len(X) > 70:
            _disp('You must introduce new markers to plot more than 70 cases.')
            _disp('The ''marker'' character array need to be extended inside the code.')
        
        if len(X) <= len(kind):
            # Define markers with specified color
            marker = []
            markercolor = []
            for color in colorm:
                for symbol in kind:
                    marker.append(symbol + option['markercolor'])
                    rgba = clr.to_rgb(option['markercolor']) + (alpha,)
                    markercolor.append(rgba)
        else:
            # Define markers and colors using predefined list
            marker = []
            markercolor = [] #Bug Fix: missing array initialization
            for color in colorm:
                for symbol in kind:
                    marker.append(symbol + color)
                    rgba = clr.to_rgb(color) + (alpha,)
                    markercolor.append(rgba)
        
        # Plot markers at data points
        limit = option['axismax']
        hp = ()
        markerlabel = []
        for i, xval in enumerate(X):
            if abs(X[i]) <= limit and abs(Y[i]) <= limit:
                h = plt.plot(X[i],Y[i],marker[i], markersize = markerSize, 
                     markerfacecolor = markercolor[i],
                     markeredgecolor = marker[i][1],
                     markeredgewidth = 2)
                hp += tuple(h)
                markerlabel.append(option['markerlabel'][i])

        # Add legend
        if len(markerlabel) == 0:
            warnings.warn('No markers within axis limit ranges.')
        else:
            add_marker_legend(markerlabel, option, rgba, markerSize, fontSize, hp)
    else:
        # Plot markers as dots of a single color with accompanying labels
        # and no legend
        
        # Plot markers at data points
        axismin = option['axismin']
        axismax = option['axismax']
        rgba = clr.to_rgb(option['markercolor']) + (alpha,) 
        for i,xval in enumerate(X):
            if X[i] >= axismin[index_x] and X[i] <= axismax[index_x] and \
                        Y[i] >= axismin[index_y] and Y[i] <= axismax[index_y]:
                # Plot marker
                marker = option['markersymbol']
                plt.plot(X[i],Y[i],marker, markersize = markerSize, 
                     markerfacecolor = rgba,
                     markeredgecolor = option['markercolor'])
                
                # Check if marker labels provided
                if type(option['markerlabel']) is list:
                    # Label marker
                    xtextpos = X[i]
                    ytextpos = Y[i]
                    plt.text(xtextpos,ytextpos,option['markerlabel'][i], 
                             color = option['markerlabelcolor'],
                             verticalalignment = 'bottom',
                             horizontalalignment = 'right',
                             fontsize = fontSize)

        # Add legend if labels provided as dictionary
        markerlabel = option['markerlabel']
        if type(markerlabel) is dict:
            add_marker_legend(markerlabel, option, rgba, markerSize, fontSize)

def _disp(text):
    print(text)



def _default_options() -> dict:
    '''
    Set default optional arguments for time_series_plot function.
    
    Sets the default optional arguments for the TIME_SERIES_PLOT 
    function in an OPTION dictionary. Default values are 
    assigned to selected optional arguments. 
    
    INPUTS:
    None
    option : dictionary containing option values. (Refer to 
             display_display_time_series_options function for more information.)
    option['alpha']           : blending of symbol face color (0.0 
                                transparent through 1.0 opaque). (Default : 1.0)
    option['axislim']         : axes limits [xmin, xmax, ymin, ymax] (Default: data values)
    option['linespec']  : line specification (default solid black, 'k-')
    option['linewidth'] : line width specification (default rcParams 'lines.linewidth')
    option['circles']         : radii of circles to draw to indicate 
                                isopleths of standard deviation (empty by default)
    option['cmapzdata']       : data values to use for color mapping of
                                markers, e.g. RMSD or BIAS. (Default empty)
    option['colormap']        : 'on'/'off' switch to map color shading of
                                 markers to CMapZData values ('on') or min to
                                 max range of CMapZData values ('off').
                                 (Default : 'on')
    option['layout']           : matrix layout for subplots specified as a list 
                                 [m,n] where m is number of rows and n number of columns.
                                 A value of m = 0 lets package specify number of rows
                                 per figure. (Default: [0,2])
    option['locationcolorbar'] : location for the colorbar, 'NorthOutside' or
                                 'EastOutside'
    option['markerdisplayed'] : markers to use for individual experiments
                                = 'line' to get a continuous line
                                = 'marker' to display only markers at provided points
    option['markerlabel']     : name of the experiment to use for marker
    option['markerlabelcolor'] : marker label color (Default 'k')
    option['markerlegend']    : 'on'/'off' switch to display marker legend
                                (Default 'off')
    option['markersize']      : marker size (Default 10)
    option['markersymbol']    : marker symbol (Default 'o')

    option['overlay']         : 'on'/'off' switch to overlay current
                                statistics on Taylor diagram (Default 'off')
                                Only markers will be displayed.
    option['plottype']        : type of x-y plot
                                'linear', standard linear plot (Default)
                                'loglog', log scaling on both the x and y axis
                                'semilogx', log scaling on the x axis
                                'semilogy', log scaling on the y axis
    option['ticks']           : define tick positions (default is that used 
                                by the axis function)
    option['titlecolorbar']   : title for the colorbar
    option['xticklabelpos']   : position of the tick labels along the x-axis 
                                (empty by default)
    option['yticklabelpos']   : position of the tick labels along the y-axis 
                                (empty by default)
    '''
    from matplotlib import rcParams

    # Set default parameters for all options
    option = {}
    option['alpha'] = 1.0
    option['axeslabel'] = []
    option['axislim'] = None
    option['axismax'] = []
    option['axismin'] = []
    option['layout'] = [0, 2]
    option['linespec'] = 'k-'
    option['linewidth'] = rcParams.get('lines.linewidth')

    option['index_x'] = None
    option['index_y'] = None

    option['markercolor'] = 'r'
    option['markerdisplayed'] = 'line'
    option['markerlabel'] = ''
    option['markerlabelcolor'] = 'k'
    option['markerlegend'] = 'off'
    option['markersize'] = 10
    option['markersymbol'] = 'o'

    option['overlay'] = 'off'
    option['plottype'] = 'linear'
    option['ticks'] = []
    option['titlecolorbar'] = ''
    option['xticklabelpos'] = []
    option['yticklabelpos'] = []
         
    return option

def _get_options(option, **kwargs) -> dict:
    '''
    Get values for optional arguments for time_series_plot function.
    
    Gets the default optional arguments for the TIME_SERIES_PLOT 
    function in an OPTION dictionary. 
    
    INPUTS:
    option  : dictionary containing default option values
    *kwargs : variable-length keyword argument list. The keywords by 
              definition are dictionaries with keys that must correspond to 
              one of the choices given in the _default_options function.
        
    OUTPUTS:
    option : dictionary containing option values
  
    Author:
    
    Peter A. Rochford
        rochford.peter1@gmail.com

    Created on Aug 8, 2023
    Revised on Aug 8, 2023
    '''    
    # Check for valid keys and values in dictionary
    for optname, optvalue in kwargs.items():
        optname = optname.lower()
        if not optname in option:
            raise ValueError('Unrecognized option: ' + optname)
        else:
            # Replace option value with that from arguments
            option[optname] = optvalue

            # Check values for specific options
            if optname == 'cmapzdata':
                if isinstance(option[optname], str):
                    raise ValueError('cmapzdata cannot be a string!')
                elif isinstance(option[optname], bool):
                    raise ValueError('cmapzdata cannot be a boolean!')
                option['cmapzdata'] = optvalue
            elif optname == 'markerlabel':
                if type(optvalue) is list:
                    option['markerlabel'] = optvalue
                elif type(optvalue) is dict:
                    option['markerlabel'] = optvalue
                else:
                    raise ValueError('markerlabel value is not a list or dictionary: ' +
                                     str(optvalue))
 
        del optname, optvalue   
    
    # Set layout if an overlay
    if option['overlay'] == 'on':
        option['layout'][0] = number_rows
        option['layout'][1] = number_cols
    else: 
        number_rows = option['layout'][0]
        number_cols = option['layout'][1] 

    return option   

def get_time_series_plot_options(**kwargs):
    '''
    Get optional arguments for time_series_plot function.
    
    Retrieves the optional arguments supplied to the time_series_plot 
    function as a variable-length keyword argument list (*KWARGS), and
    returns the values in an OPTION dictionary. Default values are 
    assigned to selected optional arguments. The function will terminate
    with an error if an unrecognized optional argument is supplied.
            
    ToDo: update as package evolves
    
    INPUTS:
    *kwargs : variable-length keyword argument list. The keywords by 
              definition are dictionaries with keys that must correspond to 
              one choices given in OUTPUTS below.
    
    OUTPUTS:
    option : dictionary containing option values. (Refer to 
             display_display_time_series_options function for more information.)
  
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com

    Created on Aug 8, 2023
    Revised on Aug 8, 2023
    '''
    global number_cols, number_rows, subplot_axis

    nargin = len(kwargs)

    # Set default parameters for all options
    option = _default_options()

    # No options requested, so return with only defaults
    if nargin == 0: return option

    # Check for valid keys and values in dictionary
    # Allows user to override options specified in CSV file
    option = _get_options(option, **kwargs)
    
    return option



import matplotlib.pyplot as plt
import matplotlib

from opendss_scivis import add_line_legend

def plot_xy_line(X,Y,option):
    '''
    Plots a one-dimensional variable in an x-y plot using lines.
    
    Plots a function Y(X) such as frequency with time.
    
    INPUTS:
    x       : independent coordinates
    y       : function values at x-coordinates
    option  : dictionary containing option values. (Refer to 
              get_time_series_plot_options function for more information.)
    option['linespec']  : line specification (default solid black, 'k-')
    option['linewidth'] : line width specification (default rcParams 'lines.linewidth')
    option['plottype']  : type of x-y plot
                          'linear', standard linear plot (Default)
                          'loglog', log scaling on both the x and y axis
                          'semilogx', log scaling on the x axis
                          'semilogy', log scaling on the y axis

    OUTPUTS:
    None
    
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com

    Created on May 1, 2022
    '''

    # Plot line according to plot type
    plottype = option['plottype']
    if plottype == "linear":
        plt.plot(X,Y, option['linespec'], linewidth=option['linewidth'])
    elif plottype == "loglog":
        plt.loglog(X,Y, option['linespec'], linewidth=option['linewidth'])
    elif plottype == "semilogx":
        plt.semilogx(X,Y, option['linespec'], linewidth=option['linewidth'])
    elif plottype == "semilogy":
        plt.semilogy(X,Y, option['linespec'], linewidth=option['linewidth'])
    else:
        raise ValueError('Unknown option: ' + plottype)
                



'''
Created on Jul 10, 2023

@author: FF16GK3
'''
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap
import numpy as np
# Create a Basemap instance
map = Basemap(projection='mill', llcrnrlat=-90, urcrnrlat=90, llcrnrlon=-180, urcrnrlon=180, resolution='c')

# Draw coastlines, countries, and boundaries
map.drawcoastlines(linewidth=0.5)
map.drawcountries(linewidth=0.5)
map.drawmapboundary(fill_color='lightblue')

# Generate random scalar field data
lats = np.linspace(-90, 90, 100)
lons = np.linspace(-180, 180, 200)
lons, lats = np.meshgrid(lons, lats)
data = np.random.rand(lats.shape[0], lons.shape[1])

# Plot scalar field data
x, y = map(lons, lats)
plt.contourf(x, y, data, cmap='viridis')

# Add colorbar and title
plt.colorbar(label='Scalar Field')
plt.title('Scalar Field Map')

# Create a Basemap instance
map = Basemap(projection='mill', llcrnrlat=-90, urcrnrlat=90, llcrnrlon=-180, urcrnrlon=180, resolution='c')

# Draw coastlines, countries, and boundaries
map.drawcoastlines(linewidth=0.5)
map.drawcountries(linewidth=0.5)
map.drawmapboundary(fill_color='lightblue')

# Generate random vector field data
lats = np.linspace(-90, 90, 20)
lons = np.linspace(-180, 180, 40)
lons, lats = np.meshgrid(lons, lats)
u = np.random.randn(lats.shape[0], lons.shape[1])
v = np.random.randn(lats.shape[0], lons.shape[1])

# Plot vector field data
x, y = map(lons, lats)
map.quiver(x, y, u, v, scale=1000, pivot='middle', color='r')

# Add title
plt.title('Vector Field Map')

# Display the map
plt.show()




from math import ceil, floor, log
import matplotlib.ticker as ticker
import numpy as np

def get_time_series_axes(x,y,option):
    '''
    Get axes value for time_series_plot function.
    
    Determines the axes information for a time series plot given the axis 
    values (X,Y) and the options in the data structure OPTION returned by 
    the get_time_series_plot_options function.
    
    INPUTS:
    x       : values for x-axis
    y       : values for y-axis
    option  : dictionary containing option values. (Refer to 
              get_time_series_plot_options function for more information.)
    option['index_x'] : index for x values in option
    option['index_y'] : index for y values in option
    
    OUTPUTS:
    axes           : dictionary containing axes information for time series plot
    axes['xtick']  : x-values at which to place tick marks
    axes['ytick']  : y-values at which to place tick marks
    axes['xlabel'] : labels for xtick values
    axes['ylabel'] : labels for ytick values
    option : dictionary containing updated option values
    
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com
    
    Created on Apr 25, 2022
    '''

    # Specify max & min for axes
    maxx, maxy = _getmax(option,x,y)
    minx, miny = _getmin(option,x,y)
    
    option['axismax'].append(maxy)

    # Check if min & max the same within a tolerance
    equal = np.allclose(miny,maxy)
    if equal:
        # set min & max to suitable range
        offset = 0.1*maxy
        miny = miny - offset
        maxy = maxy + offset

    # Determine tick values marks according to plot type
    plottype = option['plottype']
    xtickvals = None
    ytickvals = None
    if plottype == "linear":
        xtickvals = ticker.AutoLocator().tick_values(minx, maxx)
        ytickvals = ticker.AutoLocator().tick_values(miny, maxy)
    elif plottype == "loglog":
        xtickvals = [ 10**int(x) for x in np.arange(np.log10(minx), np.log10(maxx)+1)]
        # ytickvals = ticker.LogLocator(base=10).tick_values(miny, maxy)
        ytickvals = [ 10**int(y) for y in np.arange(np.log10(miny), np.log10(maxy)+1)]
        # Adjust tick values if outside specified range
        xvals = [x for x in xtickvals if x >= minx and x <= maxx]
        yvals = [y for y in ytickvals if y >= miny and y <= maxy]
        xtickvals = np.array(xvals)
        ytickvals = np.array(yvals)
    elif plottype == "semilogx":
        xtickvals = ticker.LogLocator(base=10.0).tick_values(minx, maxx)
        ytickvals = ticker.AutoLocator().tick_values(miny, maxy)
    elif plottype == "semilogy":
        xtickvals = ticker.AutoLocator().tick_values(minx, maxx)
        ytickvals = ticker.LogLocator(base=10.0).tick_values(miny, maxy)
    else:
        raise ValueError('Unknown option: ' + plottype)

    ntest = np.sum(xtickvals > 0)
    if ntest > 0:
        nxticks = len(xtickvals) - 1
        nyticks = len(ytickvals) - 1
        
        # Save nxticks and nyticks as function attributes for later 
        # retrieval in function calls
        get_time_series_axes.nxticks = nxticks
        get_time_series_axes.nyticks = nyticks
    else:
        # Use function attributes for nxticks and nyticks
        if hasattr(get_time_series_axes, 'nxticks') and \
            hasattr(get_time_series_axes, 'nxticks'):
            nxticks = get_time_series_axes.nxticks
            nyticks = get_time_series_axes.nyticks
        else:
            raise ValueError('No saved values for nxticks & nyticks.')
    
    # Convert to integer if whole number
    if type(minx) is float and minx.is_integer(): minx = int(round(minx))
    if type(miny) is float and miny.is_integer(): miny = int(round(miny))
    if type(maxx) is float and maxx.is_integer(): maxx = int(round(maxx))
    if type(maxy) is float and maxy.is_integer(): maxy = int(round(maxy))
    
    # Determine tick values
    if len(option['ticks']) > 0:
        xtick = option['ticks']
        ytick = option['ticks']
    else:
        xtick = xtickvals
        ytick = ytickvals

    # Assign tick label positions
    if len(option['xticklabelpos']) == 0:
        option['xticklabelpos'] = xtick
    if len(option['yticklabelpos']) == 0:
        option['yticklabelpos'] = ytick

    # # Set tick labels using provided tick label positions
    # xlabel =[]; ylabel = [];
    #
    # # Set x tick labels
    # for i in range(len(xtick)):
    #     index = np.where(option['xticklabelpos'] == xtick[i])
    #     if len(index) > 0:
    #         label = get_axis_tick_label(xtick[i])
    #         xlabel.append(label)
    #     else:
    #         xlabel.append('')
    #
    # # Set y tick labels
    # for i in range(len(ytick)):
    #     index = np.where(option['yticklabelpos'] == ytick[i])
    #     if len(index) > 0:
    #         label = get_axis_tick_label(ytick[i])
    #         ylabel.append(label)
    #     else:
    #         ylabel.append('')
    
    # Store output variables in data structure
    axes = {}
    axes['xtick'] = xtick
    axes['ytick'] = ytick
    # axes['xlabel'] = xlabel
    # axes['ylabel'] = ylabel
    
    return axes

def _getmax(option,x,y):
    '''
    Get maximum axes values according to plot type.
    
    INPUTS:
    x       : values for x-axis
    y       : values for y-axis
    option  : dictionary containing option values. (Refer to 
              get_time_series_plot_options function for more information.)
    option['index_x'] : index for x values in option
    option['index_y'] : index for y values in option
    option['plottype']        : type of x-y plot
                                'linear', standard linear plot (Default)
                                'loglog', log scaling on both the x and y axis
                                'semilogx', log scaling on the x axis
                                'semilogy', log scaling on the y axis
    
    OUTPUTS:
    maxx : maximum for x axis
    maxy : maximum for y axis
    '''
    index_x = option['index_x']
    index_y = option['index_y']

    plottype = option['plottype']
    if plottype == "linear":
        # Axis limit not specified
        maxx = np.amax(x)
        maxy = np.amax(y)
    elif plottype == "loglog":
        maxx = 10**floor(log(np.max(x),10))
        maxy = 10**floor(log(np.max(y),10))
    elif plottype == "semilogx":
        maxx = 10**floor(log(np.max(x),10))
        maxy = np.amax(y)
    elif plottype == "semilogy":
        maxx = np.amax(x)
        maxy = 10**floor(log(np.max(y),10))
    else:
        raise ValueError('Unknown option: ' + plottype)
        
    if option['axislim'][index_x] is not None:
        # Axis limit is specified
        maxx = option['axislim'][index_x+1]

    if option['axislim'][index_y] is not None:
        # Axis limit is specified
        maxy = option['axislim'][index_y+2]
        
    return maxx, maxy

def _getmin(option,x,y):
    '''
    Get minimum axes values according to plot type.
    
    INPUTS:
    x       : values for x-axis
    y       : values for y-axis
    option  : dictionary containing option values. (Refer to 
              get_time_series_plot_options function for more information.)
    option['index_x'] : index for x values in option
    option['index_y'] : index for y values in option
    option['plottype']        : type of x-y plot
                                'linear', standard linear plot (Default)
                                'loglog', log scaling on both the x and y axis
                                'semilogx', log scaling on the x axis
                                'semilogy', log scaling on the y axis
    
    OUTPUTS:
    minx : minimum for x axis
    miny : minimum for y axis
    '''
    index_x = option['index_x']
    index_y = option['index_y']

    plottype = option['plottype']
    if plottype == "linear":
        # Axis limit not specified
        minx = np.amin(x)
        miny = np.amin(y)
    elif plottype == "loglog":
        minx = 10**ceil(log(np.min(x),10))
        miny = 10**ceil(log(np.min(y),10))
    elif plottype == "semilogx":
        minx = 10**ceil(log(np.min(x),10))
        miny = np.amin(y)
    elif plottype == "semilogy":
        minx = np.amin(x)
        miny = 10**ceil(log(np.min(y),10))
    else:
        raise ValueError('Unknown option: ' + plottype)
        
    if option['axislim'][index_x] is not None:
        # Axis limit is specified
        minx = option['axislim'][index_x]

    if option['axislim'][index_y] is not None:
        # Axis limit is specified
        miny = option['axislim'][index_y+1]
        
    return minx, miny


import math
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

def add_line_legend(markerLabel, option, fontSize, hp = []):
    '''
    To Do: change for line plots
    Adds a legend of markers.
    
    Adds a legend to a plot according to the data type containing the 
    provided labels. If labels are provided as a list they will appear 
    in the legend beside the marker provided in the list of handles in 
    a one-to-one match. If labels are provided as a dictionary they will 
    appear beside a dot with the color value given to the label.
    
    INPUTS:
    markerLabel : list or dict variable containing markers and labels to
                  appear in legend
                  
                  A list variable must have the format:
                  markerLabel = ['M1', 'M2', 'M3']
                  
                  A dictionary variable must have the format:
                  markerLabel = = {'ERA-5': 'r', 'TRMM': 'b'}
                  where each key is the label and each value the color for 
                  the marker
    
    hp : list of plot handles that match markerLabel when latter is a list
    option : dictionary containing option values. (Refer to 
        GET_TARGET_DIAGRAM_OPTIONS function for more information.)
    option['numberpanels'] : Number of panels to display
                             = 1 for positive correlations
                             = 2 for positive and negative correlations
    rgba : a 4-tuple where the respective tuple components represent red, 
           green, blue, and alpha (opacity) values for a color
    markerSize : point size of markers
    fontSize : font size in points of labels
    
    OUTPUTS:
    None

    Created on Mar 2, 2019
    Revised on Mar 2, 2019
    
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com
    '''

    if type(markerLabel) is list:
        
        # Check for empty list of plot handles
        if len(hp) == 0:
            raise ValueError('Empty list of plot handles')
        elif len(hp) != len(markerLabel):
            raise ValueError('Number of labels and plot handle do not match: ' +
                             str(len(markerLabel)) + ' != ' + str(len(hp)))
        
        # Add legend using labels provided as list
        if len(markerLabel) <= 6:
            # Put legend in a default location
            markerlabel = tuple(markerLabel)
            plt.legend(hp, markerlabel, loc = 'upper right',
                                 fontsize = fontSize, numpoints=1,
                                 bbox_to_anchor=(1.2,1.0))
        else:
            # Put legend to right of the plot in multiple columns as needed

            nmarkers = len(markerLabel)
            ncol = int(math.ceil(nmarkers / 15.0))
            markerlabel = tuple(markerLabel)

            # Shift figure to include legend
            plt.gcf().subplots_adjust(right=0.6)

            # Plot legend of multi-column markers
            # Note: do not use bbox_to_anchor as this cuts off the legend
            plt.legend(hp, markerlabel, loc = (1.1, 0.25),
                        fontsize = fontSize, numpoints=1, ncol = ncol)

    elif type(markerLabel) is dict:
        
        # Add legend using labels provided as dictionary
            
        # Define legend elements
        legend_elements = []
        for key, value in markerLabel.items():
            legend_object = Line2D([0], [0], marker='.', markeredgecolor = value, label=key, linestyle='')
            legend_elements.append(legend_object)

        # Put legend in a default location
        plt.legend(handles=legend_elements, loc = 'upper right',
                             fontsize = fontSize, numpoints=1,
                             bbox_to_anchor=(1.2,1.0))

        if _checkKey(option, 'numberpanels') and option['numberpanels'] == 2:
            # add padding so legend is not cut off
            plt.tight_layout(pad=1)
    else:
        raise Exception('markerLabel type is not a list or dictionary: ' + 
                        str(type(markerLabel)))
    
def _checkKey(dictionary, key): 
    if key in dictionary.keys(): 
        return True
    else: 
        return False 



'''
A script to create a pip release of the OpenDSS_SciVis package.

Description of the package can be found in the wiki:
https://github.com/PeterRochford/OpenDSS_SciVis/wiki

Created on Aug 11, 2023

@author: Kevin Wu
'''
from setuptools import setup, find_packages

setup(
    name='OpenDSS_SciVis',
    version='1.0',
    packages=find_packages(),
    install_requires=[
        'matplotlib',
        'numpy',
        'pandas',
        'scipy',
        'SkillMetrics'
    ],
    author='Peter Rochford',
    author_email='peter.rochford@xatorcorp.com',
    description='A Python package for analysis and visualization of data produced by the OpenDSS software application.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/PeterRochford/OpenDSS_SciVis/tree/main',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.11',
    ],
)



import matplotlib.pyplot as plt
import opendss_scivis as osv

def plot_xy(X,Y,option,subplot_axis):
    '''
    Plots a one-dimensional variable in an x-y plot.
    
    Plots a function Y(X) such as frequency with time.
    
    INPUTS:
    x       : independent coordinates
    y       : function values at x-coordinates
    option  : dictionary containing option values. (Refer to 
              get_time_series_plot_options function for more information.)
    option['markerdisplayed'] : markers to use for plots
    subplot_axis : axis handle for subplot

    OUTPUTS:
    None
    
    Author: Peter A. Rochford
            Xator Corporation
            www.xatorcorp.com

    Created on May 1, 2022
    '''
    
    plt.axes(subplot_axis) # set axis for subplot

    lowcase = option['markerdisplayed'].lower()
    if lowcase == 'line':
        osv.plot_xy_line(X,Y,option)
    elif lowcase == 'marker':
        osv.plot_xy_markers(X,Y,option)
    else:
        raise ValueError('Unrecognized option: ' + 
                         option['markerdisplayed'])
